# coding: utf-8

# (C) Copyright IBM Corp. 2023.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# IBM OpenAPI SDK Code Generator Version: 3.64.1-cee95189-20230124-211647

"""
Billing units for IBM Cloud Enterprise

API Version: 1.0.0
"""

from datetime import datetime
from enum import Enum
from typing import Dict, List
import json

from ibm_cloud_sdk_core import BaseService, DetailedResponse, get_query_param
from ibm_cloud_sdk_core.authenticators.authenticator import Authenticator
from ibm_cloud_sdk_core.get_authenticator import get_authenticator_from_environment
from ibm_cloud_sdk_core.utils import datetime_to_string, string_to_datetime

from .common import get_sdk_headers

##############################################################################
# Service
##############################################################################


class EnterpriseBillingUnitsV1(BaseService):
    """The Enterprise Billing Units V1 service."""

    DEFAULT_SERVICE_URL = 'https://billing.cloud.ibm.com'
    DEFAULT_SERVICE_NAME = 'enterprise_billing_units'

    @classmethod
    def new_instance(
        cls,
        service_name: str = DEFAULT_SERVICE_NAME,
    ) -> 'EnterpriseBillingUnitsV1':
        """
        Return a new client for the Enterprise Billing Units service using the
               specified parameters and external configuration.
        """
        authenticator = get_authenticator_from_environment(service_name)
        service = cls(authenticator)
        service.configure_service(service_name)
        return service

    def __init__(
        self,
        authenticator: Authenticator = None,
    ) -> None:
        """
        Construct a new client for the Enterprise Billing Units service.

        :param Authenticator authenticator: The authenticator specifies the authentication mechanism.
               Get up to date information from https://github.com/IBM/python-sdk-core/blob/main/README.md
               about initializing the authenticator of your choice.
        """
        BaseService.__init__(self, service_url=self.DEFAULT_SERVICE_URL, authenticator=authenticator)

    #########################
    # Billing Units
    #########################

    def get_billing_unit(self, billing_unit_id: str, **kwargs) -> DetailedResponse:
        """
        Get billing unit by ID.

        Return the billing unit information if it exists.

        :param str billing_unit_id: The ID of the requested billing unit.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `BillingUnit` object
        """

        if not billing_unit_id:
            raise ValueError('billing_unit_id must be provided')
        headers = {}
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME, service_version='V1', operation_id='get_billing_unit'
        )
        headers.update(sdk_headers)

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        path_param_keys = ['billing_unit_id']
        path_param_values = self.encode_path_vars(billing_unit_id)
        path_param_dict = dict(zip(path_param_keys, path_param_values))
        url = '/v1/billing-units/{billing_unit_id}'.format(**path_param_dict)
        request = self.prepare_request(method='GET', url=url, headers=headers)

        response = self.send(request, **kwargs)
        return response

    def list_billing_units(
        self,
        *,
        account_id: str = None,
        enterprise_id: str = None,
        account_group_id: str = None,
        limit: int = None,
        start: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        List billing units.

        Return matching billing unit information if any exists. Omits internal properties
        and enterprise account ID from the billing unit.

        :param str account_id: (optional) The enterprise account ID.
        :param str enterprise_id: (optional) The enterprise ID.
        :param str account_group_id: (optional) The account group ID.
        :param int limit: (optional) Return results up to this limit. Valid values
               are between 0 and 100.
        :param str start: (optional) The pagination offset. This represents the
               index of the first returned result.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `BillingUnitsList` object
        """

        headers = {}
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME, service_version='V1', operation_id='list_billing_units'
        )
        headers.update(sdk_headers)

        params = {
            'account_id': account_id,
            'enterprise_id': enterprise_id,
            'account_group_id': account_group_id,
            'limit': limit,
            'start': start,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v1/billing-units'
        request = self.prepare_request(method='GET', url=url, headers=headers, params=params)

        response = self.send(request, **kwargs)
        return response

    #########################
    # Billing Options
    #########################

    def list_billing_options(
        self, billing_unit_id: str, *, limit: int = None, start: str = None, **kwargs
    ) -> DetailedResponse:
        """
        List billing options.

        Return matching billing options if any exist. Show subscriptions and promotional
        offers that are available to a billing unit.

        :param str billing_unit_id: The billing unit ID.
        :param int limit: (optional) Return results up to this limit. Valid values
               are between 0 and 100.
        :param str start: (optional) The pagination offset. This represents the
               index of the first returned result.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `BillingOptionsList` object
        """

        if not billing_unit_id:
            raise ValueError('billing_unit_id must be provided')
        headers = {}
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME, service_version='V1', operation_id='list_billing_options'
        )
        headers.update(sdk_headers)

        params = {
            'billing_unit_id': billing_unit_id,
            'limit': limit,
            'start': start,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v1/billing-options'
        request = self.prepare_request(method='GET', url=url, headers=headers, params=params)

        response = self.send(request, **kwargs)
        return response

    #########################
    # Credit Pools
    #########################

    def get_credit_pools(
        self,
        billing_unit_id: str,
        *,
        date: str = None,
        type: str = None,
        limit: int = None,
        start: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Get credit pools.

        Get credit pools for a billing unit. Credit pools can be either platform or
        support credit pools. The platform credit pool contains credit from platform
        subscriptions and promotional offers. The support credit pool contains credit from
        support subscriptions.

        :param str billing_unit_id: The ID of the billing unit.
        :param str date: (optional) The date in the format of YYYY-MM.
        :param str type: (optional) Filters the credit pool by type, either
               `PLATFORM` or `SUPPORT`.
        :param int limit: (optional) Return results up to this limit. Valid values
               are between 0 and 100.
        :param str start: (optional) The pagination offset. This represents the
               index of the first returned result.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `CreditPoolsList` object
        """

        if not billing_unit_id:
            raise ValueError('billing_unit_id must be provided')
        headers = {}
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME, service_version='V1', operation_id='get_credit_pools'
        )
        headers.update(sdk_headers)

        params = {
            'billing_unit_id': billing_unit_id,
            'date': date,
            'type': type,
            'limit': limit,
            'start': start,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v1/credit-pools'
        request = self.prepare_request(method='GET', url=url, headers=headers, params=params)

        response = self.send(request, **kwargs)
        return response


##############################################################################
# Models
##############################################################################


class BillingOption:
    """
    Information about a billing option.

    :attr str id: (optional) The ID of the billing option.
    :attr str billing_unit_id: (optional) The ID of the billing unit that's
          associated with the billing option.
    :attr datetime start_date: (optional) The start date of billing option.
    :attr datetime end_date: (optional) The end date of billing option.
    :attr str state: (optional) The state of the billing option. The valid values
          include `ACTIVE, `SUSPENDED`, and `CANCELED`.
    :attr str type: (optional) The type of billing option. The valid values are
          `SUBSCRIPTION` and `OFFER`.
    :attr str category: (optional) The category of the billing option. The valid
          values are `PLATFORM`, `SERVICE`, and `SUPPORT`.
    :attr dict payment_instrument: (optional) The payment method for support.
    :attr int duration_in_months: (optional) The duration of the billing options in
          months.
    :attr int line_item_id: (optional) The line item ID for support.
    :attr dict billing_system: (optional) The support billing system.
    :attr str renewal_mode_code: (optional) The renewal code for support. This code
          denotes whether the subscription automatically renews, is assessed monthly, and
          so on.
    :attr datetime updated_at: (optional) The date when the billing option was
          updated.
    """

    def __init__(
        self,
        *,
        id: str = None,
        billing_unit_id: str = None,
        start_date: datetime = None,
        end_date: datetime = None,
        state: str = None,
        type: str = None,
        category: str = None,
        payment_instrument: dict = None,
        duration_in_months: int = None,
        line_item_id: int = None,
        billing_system: dict = None,
        renewal_mode_code: str = None,
        updated_at: datetime = None,
    ) -> None:
        """
        Initialize a BillingOption object.

        :param str id: (optional) The ID of the billing option.
        :param str billing_unit_id: (optional) The ID of the billing unit that's
               associated with the billing option.
        :param datetime start_date: (optional) The start date of billing option.
        :param datetime end_date: (optional) The end date of billing option.
        :param str state: (optional) The state of the billing option. The valid
               values include `ACTIVE, `SUSPENDED`, and `CANCELED`.
        :param str type: (optional) The type of billing option. The valid values
               are `SUBSCRIPTION` and `OFFER`.
        :param str category: (optional) The category of the billing option. The
               valid values are `PLATFORM`, `SERVICE`, and `SUPPORT`.
        :param dict payment_instrument: (optional) The payment method for support.
        :param int duration_in_months: (optional) The duration of the billing
               options in months.
        :param int line_item_id: (optional) The line item ID for support.
        :param dict billing_system: (optional) The support billing system.
        :param str renewal_mode_code: (optional) The renewal code for support. This
               code denotes whether the subscription automatically renews, is assessed
               monthly, and so on.
        :param datetime updated_at: (optional) The date when the billing option was
               updated.
        """
        self.id = id
        self.billing_unit_id = billing_unit_id
        self.start_date = start_date
        self.end_date = end_date
        self.state = state
        self.type = type
        self.category = category
        self.payment_instrument = payment_instrument
        self.duration_in_months = duration_in_months
        self.line_item_id = line_item_id
        self.billing_system = billing_system
        self.renewal_mode_code = renewal_mode_code
        self.updated_at = updated_at

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'BillingOption':
        """Initialize a BillingOption object from a json dictionary."""
        args = {}
        if 'id' in _dict:
            args['id'] = _dict.get('id')
        if 'billing_unit_id' in _dict:
            args['billing_unit_id'] = _dict.get('billing_unit_id')
        if 'start_date' in _dict:
            args['start_date'] = string_to_datetime(_dict.get('start_date'))
        if 'end_date' in _dict:
            args['end_date'] = string_to_datetime(_dict.get('end_date'))
        if 'state' in _dict:
            args['state'] = _dict.get('state')
        if 'type' in _dict:
            args['type'] = _dict.get('type')
        if 'category' in _dict:
            args['category'] = _dict.get('category')
        if 'payment_instrument' in _dict:
            args['payment_instrument'] = _dict.get('payment_instrument')
        if 'duration_in_months' in _dict:
            args['duration_in_months'] = _dict.get('duration_in_months')
        if 'line_item_id' in _dict:
            args['line_item_id'] = _dict.get('line_item_id')
        if 'billing_system' in _dict:
            args['billing_system'] = _dict.get('billing_system')
        if 'renewal_mode_code' in _dict:
            args['renewal_mode_code'] = _dict.get('renewal_mode_code')
        if 'updated_at' in _dict:
            args['updated_at'] = string_to_datetime(_dict.get('updated_at'))
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a BillingOption object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'id') and self.id is not None:
            _dict['id'] = self.id
        if hasattr(self, 'billing_unit_id') and self.billing_unit_id is not None:
            _dict['billing_unit_id'] = self.billing_unit_id
        if hasattr(self, 'start_date') and self.start_date is not None:
            _dict['start_date'] = datetime_to_string(self.start_date)
        if hasattr(self, 'end_date') and self.end_date is not None:
            _dict['end_date'] = datetime_to_string(self.end_date)
        if hasattr(self, 'state') and self.state is not None:
            _dict['state'] = self.state
        if hasattr(self, 'type') and self.type is not None:
            _dict['type'] = self.type
        if hasattr(self, 'category') and self.category is not None:
            _dict['category'] = self.category
        if hasattr(self, 'payment_instrument') and self.payment_instrument is not None:
            _dict['payment_instrument'] = self.payment_instrument
        if hasattr(self, 'duration_in_months') and self.duration_in_months is not None:
            _dict['duration_in_months'] = self.duration_in_months
        if hasattr(self, 'line_item_id') and self.line_item_id is not None:
            _dict['line_item_id'] = self.line_item_id
        if hasattr(self, 'billing_system') and self.billing_system is not None:
            _dict['billing_system'] = self.billing_system
        if hasattr(self, 'renewal_mode_code') and self.renewal_mode_code is not None:
            _dict['renewal_mode_code'] = self.renewal_mode_code
        if hasattr(self, 'updated_at') and self.updated_at is not None:
            _dict['updated_at'] = datetime_to_string(self.updated_at)
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this BillingOption object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'BillingOption') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'BillingOption') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other

    class StateEnum(str, Enum):
        """
        The state of the billing option. The valid values include `ACTIVE, `SUSPENDED`,
        and `CANCELED`.
        """

        ACTIVE = 'ACTIVE'
        SUSPENDED = 'SUSPENDED'
        CANCELED = 'CANCELED'

    class TypeEnum(str, Enum):
        """
        The type of billing option. The valid values are `SUBSCRIPTION` and `OFFER`.
        """

        SUBSCRIPTION = 'SUBSCRIPTION'
        OFFER = 'OFFER'

    class CategoryEnum(str, Enum):
        """
        The category of the billing option. The valid values are `PLATFORM`, `SERVICE`,
        and `SUPPORT`.
        """

        PLATFORM = 'PLATFORM'
        SERVICE = 'SERVICE'
        SUPPORT = 'SUPPORT'


class BillingOptionsList:
    """
    A search result containing zero or more billing options.

    :attr int rows_count: (optional) A count of the billing units that were found by
          the query.
    :attr str next_url: (optional) Bookmark URL to query for next batch of billing
          units. This returns `null` if no additional pages are required.
    :attr List[BillingOption] resources: (optional) A list of billing units found.
    """

    def __init__(
        self, *, rows_count: int = None, next_url: str = None, resources: List['BillingOption'] = None
    ) -> None:
        """
        Initialize a BillingOptionsList object.

        :param int rows_count: (optional) A count of the billing units that were
               found by the query.
        :param str next_url: (optional) Bookmark URL to query for next batch of
               billing units. This returns `null` if no additional pages are required.
        :param List[BillingOption] resources: (optional) A list of billing units
               found.
        """
        self.rows_count = rows_count
        self.next_url = next_url
        self.resources = resources

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'BillingOptionsList':
        """Initialize a BillingOptionsList object from a json dictionary."""
        args = {}
        if 'rows_count' in _dict:
            args['rows_count'] = _dict.get('rows_count')
        if 'next_url' in _dict:
            args['next_url'] = _dict.get('next_url')
        if 'resources' in _dict:
            args['resources'] = [BillingOption.from_dict(v) for v in _dict.get('resources')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a BillingOptionsList object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'rows_count') and self.rows_count is not None:
            _dict['rows_count'] = self.rows_count
        if hasattr(self, 'next_url') and self.next_url is not None:
            _dict['next_url'] = self.next_url
        if hasattr(self, 'resources') and self.resources is not None:
            resources_list = []
            for v in self.resources:
                if isinstance(v, dict):
                    resources_list.append(v)
                else:
                    resources_list.append(v.to_dict())
            _dict['resources'] = resources_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this BillingOptionsList object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'BillingOptionsList') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'BillingOptionsList') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class BillingUnit:
    """
    Information about a billing unit.

    :attr str id: (optional) The ID of the billing unit, which is a globally unique
          identifier (GUID).
    :attr str crn: (optional) The Cloud Resource Name (CRN) of the billing unit,
          scoped to the enterprise account ID.
    :attr str name: (optional) The name of the billing unit.
    :attr str enterprise_id: (optional) The ID of the enterprise to which the
          billing unit is associated.
    :attr str currency_code: (optional) The currency code for the billing unit.
    :attr str country_code: (optional) The country code for the billing unit.
    :attr bool master: (optional) A flag that indicates whether this billing unit is
          the primary billing mechanism for the enterprise.
    :attr datetime created_at: (optional) The creation date of the billing unit.
    """

    def __init__(
        self,
        *,
        id: str = None,
        crn: str = None,
        name: str = None,
        enterprise_id: str = None,
        currency_code: str = None,
        country_code: str = None,
        master: bool = None,
        created_at: datetime = None,
    ) -> None:
        """
        Initialize a BillingUnit object.

        :param str id: (optional) The ID of the billing unit, which is a globally
               unique identifier (GUID).
        :param str crn: (optional) The Cloud Resource Name (CRN) of the billing
               unit, scoped to the enterprise account ID.
        :param str name: (optional) The name of the billing unit.
        :param str enterprise_id: (optional) The ID of the enterprise to which the
               billing unit is associated.
        :param str currency_code: (optional) The currency code for the billing
               unit.
        :param str country_code: (optional) The country code for the billing unit.
        :param bool master: (optional) A flag that indicates whether this billing
               unit is the primary billing mechanism for the enterprise.
        :param datetime created_at: (optional) The creation date of the billing
               unit.
        """
        self.id = id
        self.crn = crn
        self.name = name
        self.enterprise_id = enterprise_id
        self.currency_code = currency_code
        self.country_code = country_code
        self.master = master
        self.created_at = created_at

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'BillingUnit':
        """Initialize a BillingUnit object from a json dictionary."""
        args = {}
        if 'id' in _dict:
            args['id'] = _dict.get('id')
        if 'crn' in _dict:
            args['crn'] = _dict.get('crn')
        if 'name' in _dict:
            args['name'] = _dict.get('name')
        if 'enterprise_id' in _dict:
            args['enterprise_id'] = _dict.get('enterprise_id')
        if 'currency_code' in _dict:
            args['currency_code'] = _dict.get('currency_code')
        if 'country_code' in _dict:
            args['country_code'] = _dict.get('country_code')
        if 'master' in _dict:
            args['master'] = _dict.get('master')
        if 'created_at' in _dict:
            args['created_at'] = string_to_datetime(_dict.get('created_at'))
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a BillingUnit object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'id') and self.id is not None:
            _dict['id'] = self.id
        if hasattr(self, 'crn') and self.crn is not None:
            _dict['crn'] = self.crn
        if hasattr(self, 'name') and self.name is not None:
            _dict['name'] = self.name
        if hasattr(self, 'enterprise_id') and self.enterprise_id is not None:
            _dict['enterprise_id'] = self.enterprise_id
        if hasattr(self, 'currency_code') and self.currency_code is not None:
            _dict['currency_code'] = self.currency_code
        if hasattr(self, 'country_code') and self.country_code is not None:
            _dict['country_code'] = self.country_code
        if hasattr(self, 'master') and self.master is not None:
            _dict['master'] = self.master
        if hasattr(self, 'created_at') and self.created_at is not None:
            _dict['created_at'] = datetime_to_string(self.created_at)
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this BillingUnit object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'BillingUnit') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'BillingUnit') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class BillingUnitsList:
    """
    A search result contining zero or more billing units.

    :attr int rows_count: (optional) A count of the billing units that were found by
          the query.
    :attr str next_url: (optional) Bookmark URL to query for next batch of billing
          units. This returns `null` if no additional pages are required.
    :attr List[BillingUnit] resources: (optional) A list of billing units found.
    """

    def __init__(self, *, rows_count: int = None, next_url: str = None, resources: List['BillingUnit'] = None) -> None:
        """
        Initialize a BillingUnitsList object.

        :param int rows_count: (optional) A count of the billing units that were
               found by the query.
        :param str next_url: (optional) Bookmark URL to query for next batch of
               billing units. This returns `null` if no additional pages are required.
        :param List[BillingUnit] resources: (optional) A list of billing units
               found.
        """
        self.rows_count = rows_count
        self.next_url = next_url
        self.resources = resources

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'BillingUnitsList':
        """Initialize a BillingUnitsList object from a json dictionary."""
        args = {}
        if 'rows_count' in _dict:
            args['rows_count'] = _dict.get('rows_count')
        if 'next_url' in _dict:
            args['next_url'] = _dict.get('next_url')
        if 'resources' in _dict:
            args['resources'] = [BillingUnit.from_dict(v) for v in _dict.get('resources')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a BillingUnitsList object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'rows_count') and self.rows_count is not None:
            _dict['rows_count'] = self.rows_count
        if hasattr(self, 'next_url') and self.next_url is not None:
            _dict['next_url'] = self.next_url
        if hasattr(self, 'resources') and self.resources is not None:
            resources_list = []
            for v in self.resources:
                if isinstance(v, dict):
                    resources_list.append(v)
                else:
                    resources_list.append(v.to_dict())
            _dict['resources'] = resources_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this BillingUnitsList object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'BillingUnitsList') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'BillingUnitsList') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class CreditPool:
    """
    The credit pool for a billing unit.

    :attr str type: (optional) The type of credit, either `PLATFORM` or `SUPPORT`.
    :attr str currency_code: (optional) The currency code of the associated billing
          unit.
    :attr str billing_unit_id: (optional) The ID of the billing unit that's
          associated with the credit pool. This value is a globally unique identifier
          (GUID).
    :attr List[TermCredits] term_credits: (optional) A list of active subscription
          terms available within a credit pool.
    :attr CreditPoolOverage overage: (optional) Overage that was generated on the
          credit pool.
    """

    def __init__(
        self,
        *,
        type: str = None,
        currency_code: str = None,
        billing_unit_id: str = None,
        term_credits: List['TermCredits'] = None,
        overage: 'CreditPoolOverage' = None,
    ) -> None:
        """
        Initialize a CreditPool object.

        :param str type: (optional) The type of credit, either `PLATFORM` or
               `SUPPORT`.
        :param str currency_code: (optional) The currency code of the associated
               billing unit.
        :param str billing_unit_id: (optional) The ID of the billing unit that's
               associated with the credit pool. This value is a globally unique identifier
               (GUID).
        :param List[TermCredits] term_credits: (optional) A list of active
               subscription terms available within a credit pool.
        :param CreditPoolOverage overage: (optional) Overage that was generated on
               the credit pool.
        """
        self.type = type
        self.currency_code = currency_code
        self.billing_unit_id = billing_unit_id
        self.term_credits = term_credits
        self.overage = overage

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'CreditPool':
        """Initialize a CreditPool object from a json dictionary."""
        args = {}
        if 'type' in _dict:
            args['type'] = _dict.get('type')
        if 'currency_code' in _dict:
            args['currency_code'] = _dict.get('currency_code')
        if 'billing_unit_id' in _dict:
            args['billing_unit_id'] = _dict.get('billing_unit_id')
        if 'term_credits' in _dict:
            args['term_credits'] = [TermCredits.from_dict(v) for v in _dict.get('term_credits')]
        if 'overage' in _dict:
            args['overage'] = CreditPoolOverage.from_dict(_dict.get('overage'))
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a CreditPool object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'type') and self.type is not None:
            _dict['type'] = self.type
        if hasattr(self, 'currency_code') and self.currency_code is not None:
            _dict['currency_code'] = self.currency_code
        if hasattr(self, 'billing_unit_id') and self.billing_unit_id is not None:
            _dict['billing_unit_id'] = self.billing_unit_id
        if hasattr(self, 'term_credits') and self.term_credits is not None:
            term_credits_list = []
            for v in self.term_credits:
                if isinstance(v, dict):
                    term_credits_list.append(v)
                else:
                    term_credits_list.append(v.to_dict())
            _dict['term_credits'] = term_credits_list
        if hasattr(self, 'overage') and self.overage is not None:
            if isinstance(self.overage, dict):
                _dict['overage'] = self.overage
            else:
                _dict['overage'] = self.overage.to_dict()
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this CreditPool object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'CreditPool') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'CreditPool') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other

    class TypeEnum(str, Enum):
        """
        The type of credit, either `PLATFORM` or `SUPPORT`.
        """

        PLATFORM = 'PLATFORM'
        SUPPORT = 'SUPPORT'


class CreditPoolOverage:
    """
    Overage that was generated on the credit pool.

    :attr float cost: (optional) The number of credits used as overage.
    :attr List[dict] resources: (optional) A list of resources that generated
          overage.
    """

    def __init__(self, *, cost: float = None, resources: List[dict] = None) -> None:
        """
        Initialize a CreditPoolOverage object.

        :param float cost: (optional) The number of credits used as overage.
        :param List[dict] resources: (optional) A list of resources that generated
               overage.
        """
        self.cost = cost
        self.resources = resources

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'CreditPoolOverage':
        """Initialize a CreditPoolOverage object from a json dictionary."""
        args = {}
        if 'cost' in _dict:
            args['cost'] = _dict.get('cost')
        if 'resources' in _dict:
            args['resources'] = _dict.get('resources')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a CreditPoolOverage object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'cost') and self.cost is not None:
            _dict['cost'] = self.cost
        if hasattr(self, 'resources') and self.resources is not None:
            _dict['resources'] = self.resources
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this CreditPoolOverage object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'CreditPoolOverage') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'CreditPoolOverage') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class CreditPoolsList:
    """
    A search result containing zero or more credit pools.

    :attr int rows_count: (optional) The number of credit pools that were found by
          the query.
    :attr str next_url: (optional) A bookmark URL to the query for the next batch of
          billing units. Use a value of `null` if no additional pages are required.
    :attr List[CreditPool] resources: (optional) A list of credit pools found by the
          query.
    """

    def __init__(self, *, rows_count: int = None, next_url: str = None, resources: List['CreditPool'] = None) -> None:
        """
        Initialize a CreditPoolsList object.

        :param int rows_count: (optional) The number of credit pools that were
               found by the query.
        :param str next_url: (optional) A bookmark URL to the query for the next
               batch of billing units. Use a value of `null` if no additional pages are
               required.
        :param List[CreditPool] resources: (optional) A list of credit pools found
               by the query.
        """
        self.rows_count = rows_count
        self.next_url = next_url
        self.resources = resources

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'CreditPoolsList':
        """Initialize a CreditPoolsList object from a json dictionary."""
        args = {}
        if 'rows_count' in _dict:
            args['rows_count'] = _dict.get('rows_count')
        if 'next_url' in _dict:
            args['next_url'] = _dict.get('next_url')
        if 'resources' in _dict:
            args['resources'] = [CreditPool.from_dict(v) for v in _dict.get('resources')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a CreditPoolsList object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'rows_count') and self.rows_count is not None:
            _dict['rows_count'] = self.rows_count
        if hasattr(self, 'next_url') and self.next_url is not None:
            _dict['next_url'] = self.next_url
        if hasattr(self, 'resources') and self.resources is not None:
            resources_list = []
            for v in self.resources:
                if isinstance(v, dict):
                    resources_list.append(v)
                else:
                    resources_list.append(v.to_dict())
            _dict['resources'] = resources_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this CreditPoolsList object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'CreditPoolsList') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'CreditPoolsList') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class TermCredits:
    """
    The subscription term that is active in the current month.

    :attr str billing_option_id: (optional) The ID of the billing option from which
          the subscription term is derived.
    :attr str category: (optional) The category of the credit pool. The valid values
          are `PLATFORM`, `OFFER`, or `SERVICE` for platform credit and `SUPPORT` for
          support credit.
    :attr datetime start_date: (optional) The start date of the term in ISO format.
    :attr datetime end_date: (optional) The end date of the term in ISO format.
    :attr float total_credits: (optional) The total credit available in this term.
    :attr float starting_balance: (optional) The balance of available credit at the
          start of the current month.
    :attr float used_credits: (optional) The amount of credit used during the
          current month.
    :attr float current_balance: (optional) The balance of remaining credit in the
          subscription term.
    :attr List[dict] resources: (optional) A list of resources that used credit
          during the month.
    """

    def __init__(
        self,
        *,
        billing_option_id: str = None,
        category: str = None,
        start_date: datetime = None,
        end_date: datetime = None,
        total_credits: float = None,
        starting_balance: float = None,
        used_credits: float = None,
        current_balance: float = None,
        resources: List[dict] = None,
    ) -> None:
        """
        Initialize a TermCredits object.

        :param str billing_option_id: (optional) The ID of the billing option from
               which the subscription term is derived.
        :param str category: (optional) The category of the credit pool. The valid
               values are `PLATFORM`, `OFFER`, or `SERVICE` for platform credit and
               `SUPPORT` for support credit.
        :param datetime start_date: (optional) The start date of the term in ISO
               format.
        :param datetime end_date: (optional) The end date of the term in ISO
               format.
        :param float total_credits: (optional) The total credit available in this
               term.
        :param float starting_balance: (optional) The balance of available credit
               at the start of the current month.
        :param float used_credits: (optional) The amount of credit used during the
               current month.
        :param float current_balance: (optional) The balance of remaining credit in
               the subscription term.
        :param List[dict] resources: (optional) A list of resources that used
               credit during the month.
        """
        self.billing_option_id = billing_option_id
        self.category = category
        self.start_date = start_date
        self.end_date = end_date
        self.total_credits = total_credits
        self.starting_balance = starting_balance
        self.used_credits = used_credits
        self.current_balance = current_balance
        self.resources = resources

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'TermCredits':
        """Initialize a TermCredits object from a json dictionary."""
        args = {}
        if 'billing_option_id' in _dict:
            args['billing_option_id'] = _dict.get('billing_option_id')
        if 'category' in _dict:
            args['category'] = _dict.get('category')
        if 'start_date' in _dict:
            args['start_date'] = string_to_datetime(_dict.get('start_date'))
        if 'end_date' in _dict:
            args['end_date'] = string_to_datetime(_dict.get('end_date'))
        if 'total_credits' in _dict:
            args['total_credits'] = _dict.get('total_credits')
        if 'starting_balance' in _dict:
            args['starting_balance'] = _dict.get('starting_balance')
        if 'used_credits' in _dict:
            args['used_credits'] = _dict.get('used_credits')
        if 'current_balance' in _dict:
            args['current_balance'] = _dict.get('current_balance')
        if 'resources' in _dict:
            args['resources'] = _dict.get('resources')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a TermCredits object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'billing_option_id') and self.billing_option_id is not None:
            _dict['billing_option_id'] = self.billing_option_id
        if hasattr(self, 'category') and self.category is not None:
            _dict['category'] = self.category
        if hasattr(self, 'start_date') and self.start_date is not None:
            _dict['start_date'] = datetime_to_string(self.start_date)
        if hasattr(self, 'end_date') and self.end_date is not None:
            _dict['end_date'] = datetime_to_string(self.end_date)
        if hasattr(self, 'total_credits') and self.total_credits is not None:
            _dict['total_credits'] = self.total_credits
        if hasattr(self, 'starting_balance') and self.starting_balance is not None:
            _dict['starting_balance'] = self.starting_balance
        if hasattr(self, 'used_credits') and self.used_credits is not None:
            _dict['used_credits'] = self.used_credits
        if hasattr(self, 'current_balance') and self.current_balance is not None:
            _dict['current_balance'] = self.current_balance
        if hasattr(self, 'resources') and self.resources is not None:
            _dict['resources'] = self.resources
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this TermCredits object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'TermCredits') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'TermCredits') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other

    class CategoryEnum(str, Enum):
        """
        The category of the credit pool. The valid values are `PLATFORM`, `OFFER`, or
        `SERVICE` for platform credit and `SUPPORT` for support credit.
        """

        PLATFORM = 'PLATFORM'
        OFFER = 'OFFER'
        SERVICE = 'SERVICE'
        SUPPORT = 'SUPPORT'


##############################################################################
# Pagers
##############################################################################


class BillingUnitsPager:
    """
    BillingUnitsPager can be used to simplify the use of the "list_billing_units" method.
    """

    def __init__(
        self,
        *,
        client: EnterpriseBillingUnitsV1,
        account_id: str = None,
        enterprise_id: str = None,
        account_group_id: str = None,
        limit: int = None,
    ) -> None:
        """
        Initialize a BillingUnitsPager object.
        :param str account_id: (optional) The enterprise account ID.
        :param str enterprise_id: (optional) The enterprise ID.
        :param str account_group_id: (optional) The account group ID.
        :param int limit: (optional) Return results up to this limit. Valid values
               are between 0 and 100.
        """
        self._has_next = True
        self._client = client
        self._page_context = {'next': None}
        self._account_id = account_id
        self._enterprise_id = enterprise_id
        self._account_group_id = account_group_id
        self._limit = limit

    def has_next(self) -> bool:
        """
        Returns true if there are potentially more results to be retrieved.
        """
        return self._has_next

    def get_next(self) -> List[dict]:
        """
        Returns the next page of results.
        :return: A List[dict], where each element is a dict that represents an instance of BillingUnit.
        :rtype: List[dict]
        """
        if not self.has_next():
            raise StopIteration(message='No more results available')

        result = self._client.list_billing_units(
            account_id=self._account_id,
            enterprise_id=self._enterprise_id,
            account_group_id=self._account_group_id,
            limit=self._limit,
            start=self._page_context.get('next'),
        ).get_result()

        next = None
        next_page_link = result.get('next_url')
        if next_page_link is not None:
            next = get_query_param(next_page_link.rstrip('&'), 'start')
        self._page_context['next'] = next
        if next is None:
            self._has_next = False

        return result.get('resources')

    def get_all(self) -> List[dict]:
        """
        Returns all results by invoking get_next() repeatedly
        until all pages of results have been retrieved.
        :return: A List[dict], where each element is a dict that represents an instance of BillingUnit.
        :rtype: List[dict]
        """
        results = []
        while self.has_next():
            next_page = self.get_next()
            results.extend(next_page)
        return results


class BillingOptionsPager:
    """
    BillingOptionsPager can be used to simplify the use of the "list_billing_options" method.
    """

    def __init__(
        self,
        *,
        client: EnterpriseBillingUnitsV1,
        billing_unit_id: str,
        limit: int = None,
    ) -> None:
        """
        Initialize a BillingOptionsPager object.
        :param str billing_unit_id: The billing unit ID.
        :param int limit: (optional) Return results up to this limit. Valid values
               are between 0 and 100.
        """
        self._has_next = True
        self._client = client
        self._page_context = {'next': None}
        self._billing_unit_id = billing_unit_id
        self._limit = limit

    def has_next(self) -> bool:
        """
        Returns true if there are potentially more results to be retrieved.
        """
        return self._has_next

    def get_next(self) -> List[dict]:
        """
        Returns the next page of results.
        :return: A List[dict], where each element is a dict that represents an instance of BillingOption.
        :rtype: List[dict]
        """
        if not self.has_next():
            raise StopIteration(message='No more results available')

        result = self._client.list_billing_options(
            billing_unit_id=self._billing_unit_id,
            limit=self._limit,
            start=self._page_context.get('next'),
        ).get_result()

        next = None
        next_page_link = result.get('next_url')
        if next_page_link is not None:
            next = get_query_param(next_page_link.rstrip('&'), 'start')
        self._page_context['next'] = next
        if next is None:
            self._has_next = False

        return result.get('resources')

    def get_all(self) -> List[dict]:
        """
        Returns all results by invoking get_next() repeatedly
        until all pages of results have been retrieved.
        :return: A List[dict], where each element is a dict that represents an instance of BillingOption.
        :rtype: List[dict]
        """
        results = []
        while self.has_next():
            next_page = self.get_next()
            results.extend(next_page)
        return results



# coding: utf-8

# (C) Copyright IBM Corp. 2023.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# IBM OpenAPI SDK Code Generator Version: 3.79.0-2eb6af3d-20230905-174838

"""
Manage your tags with the Tagging API in IBM Cloud. You can attach, detach, delete, or
list all of the tags in your billing account with the Tagging API. The tag name must be
unique within a billing account. You can create tags in two formats: `key:value` or
`label`. The tagging API supports three types of tag: `user` `service`, and `access` tags.
`service` tags cannot be attached to IMS resources. `service` tags must be in the form
`service_prefix:tag_label` where `service_prefix` identifies the Service owning the tag.
`access` tags cannot be attached to IMS and Cloud Foundry resources. They must be in the
form `key:value`.

API Version: 1.2.0
"""

from enum import Enum
from typing import Dict, List
import json

from ibm_cloud_sdk_core import BaseService, DetailedResponse
from ibm_cloud_sdk_core.authenticators.authenticator import Authenticator
from ibm_cloud_sdk_core.get_authenticator import get_authenticator_from_environment
from ibm_cloud_sdk_core.utils import convert_list, convert_model

from .common import get_sdk_headers

##############################################################################
# Service
##############################################################################


class GlobalTaggingV1(BaseService):
    """The global_tagging V1 service."""

    DEFAULT_SERVICE_URL = 'https://tags.global-search-tagging.cloud.ibm.com'
    DEFAULT_SERVICE_NAME = 'global_tagging'

    @classmethod
    def new_instance(
        cls,
        service_name: str = DEFAULT_SERVICE_NAME,
    ) -> 'GlobalTaggingV1':
        """
        Return a new client for the global_tagging service using the specified
               parameters and external configuration.
        """
        authenticator = get_authenticator_from_environment(service_name)
        service = cls(authenticator)
        service.configure_service(service_name)
        return service

    def __init__(
        self,
        authenticator: Authenticator = None,
    ) -> None:
        """
        Construct a new client for the global_tagging service.

        :param Authenticator authenticator: The authenticator specifies the authentication mechanism.
               Get up to date information from https://github.com/IBM/python-sdk-core/blob/main/README.md
               about initializing the authenticator of your choice.
        """
        BaseService.__init__(self, service_url=self.DEFAULT_SERVICE_URL, authenticator=authenticator)

    #########################
    # tags
    #########################

    def list_tags(
        self,
        *,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        impersonate_user: str = None,
        account_id: str = None,
        tag_type: str = None,
        full_data: bool = None,
        providers: List[str] = None,
        attached_to: str = None,
        offset: int = None,
        limit: int = None,
        timeout: int = None,
        order_by_name: str = None,
        attached_only: bool = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Get all tags.

        Lists all tags that are in a billing account. Use the `attached_to` parameter to
        return the list of tags that are attached to the specified resource.

        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param str impersonate_user: (optional) The user on whose behalf the get
               operation must be performed (_for administrators only_).
        :param str account_id: (optional) The ID of the billing account to list the
               tags for. If it is not set, then it is taken from the authorization token.
               This parameter is required if `tag_type` is set to `service`.
        :param str tag_type: (optional) The type of the tag you want to list.
               Supported values are `user`, `service` and `access`.
        :param bool full_data: (optional) If set to `true`, this query returns the
               provider, `ghost`, `ims` or `ghost,ims`, where the tag exists and the
               number of attached resources.
        :param List[str] providers: (optional) Select a provider. Supported values
               are `ghost` and `ims`. To list both Global Search and Tagging tags and
               infrastructure tags, use `ghost,ims`. `service` and `access` tags can only
               be attached to resources that are onboarded to Global Search and Tagging,
               so you should not set this parameter to list them.
        :param str attached_to: (optional) If you want to return only the list of
               tags that are attached to a specified resource, pass the ID of the resource
               on this parameter. For resources that are onboarded to Global Search and
               Tagging, the resource ID is the CRN; for IMS resources, it is the IMS ID.
               When using this parameter, you must specify the appropriate provider (`ims`
               or `ghost`).
        :param int offset: (optional) The offset is the index of the item from
               which you want to start returning data from.
        :param int limit: (optional) The number of tags to return.
        :param int timeout: (optional) The timeout in milliseconds, bounds the
               request to run within the specified time value. It returns the accumulated
               results until time runs out.
        :param str order_by_name: (optional) Order the output by tag name.
        :param bool attached_only: (optional) Filter on attached tags. If `true`,
               it returns only tags that are attached to one or more resources. If
               `false`, it returns all tags.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `TagList` object
        """

        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='list_tags',
        )
        headers.update(sdk_headers)

        params = {
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
            'full_data': full_data,
            'providers': convert_list(providers),
            'attached_to': attached_to,
            'offset': offset,
            'limit': limit,
            'timeout': timeout,
            'order_by_name': order_by_name,
            'attached_only': attached_only,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v3/tags'
        request = self.prepare_request(
            method='GET',
            url=url,
            headers=headers,
            params=params,
        )

        response = self.send(request, **kwargs)
        return response

    def create_tag(
        self,
        tag_names: List[str],
        *,
        impersonate_user: str = None,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        account_id: str = None,
        tag_type: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Create an access management tag.

        Create an access management tag. To create an `access` tag, you must have the
        access listed in the [Granting users access to tag
        resources](https://cloud.ibm.com/docs/account?topic=account-access) documentation.
        `service` and `user` tags cannot be created upfront. They are created when they
        are attached for the first time to a resource.

        :param List[str] tag_names: An array of tag names to create.
        :param str impersonate_user: (optional) The user on whose behalf the create
               operation must be performed (_for administrators only_).
        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param str account_id: (optional) The ID of the billing account where the
               tag must be created. It is a required parameter if `impersonate_user` is
               set.
        :param str tag_type: (optional) The type of the tags you want to create.
               The only allowed value is `access`.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `CreateTagResults` object
        """

        if tag_names is None:
            raise ValueError('tag_names must be provided')
        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='create_tag',
        )
        headers.update(sdk_headers)

        params = {
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
        }

        data = {
            'tag_names': tag_names,
        }
        data = {k: v for (k, v) in data.items() if v is not None}
        data = json.dumps(data)
        headers['content-type'] = 'application/json'

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v3/tags'
        request = self.prepare_request(
            method='POST',
            url=url,
            headers=headers,
            params=params,
            data=data,
        )

        response = self.send(request, **kwargs)
        return response

    def delete_tag_all(
        self,
        *,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        providers: str = None,
        impersonate_user: str = None,
        account_id: str = None,
        tag_type: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Delete all unused tags.

        Delete the tags that are not attached to any resource.

        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param str providers: (optional) Select a provider. Supported values are
               `ghost` and `ims`.
        :param str impersonate_user: (optional) The user on whose behalf the delete
               all operation must be performed (_for administrators only_).
        :param str account_id: (optional) The ID of the billing account to delete
               the tags for. If it is not set, then it is taken from the authorization
               token. It is a required parameter if `tag_type` is set to `service`.
        :param str tag_type: (optional) The type of the tag. Supported values are
               `user`, `service` and `access`. `service` and `access` are not supported
               for IMS resources (`providers` parameter set to `ims`).
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `DeleteTagsResult` object
        """

        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='delete_tag_all',
        )
        headers.update(sdk_headers)

        params = {
            'providers': providers,
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v3/tags'
        request = self.prepare_request(
            method='DELETE',
            url=url,
            headers=headers,
            params=params,
        )

        response = self.send(request, **kwargs)
        return response

    def delete_tag(
        self,
        tag_name: str,
        *,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        providers: List[str] = None,
        impersonate_user: str = None,
        account_id: str = None,
        tag_type: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Delete an unused tag.

        Delete an existing tag. A tag can be deleted only if it is not attached to any
        resource.

        :param str tag_name: The name of tag to be deleted.
        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param List[str] providers: (optional) Select a provider. Supported values
               are `ghost` and `ims`. To delete tags both in Global Search and Tagging and
               in IMS, use `ghost,ims`.
        :param str impersonate_user: (optional) The user on whose behalf the delete
               operation must be performed (_for administrators only_).
        :param str account_id: (optional) The ID of the billing account to delete
               the tag for. It is a required parameter if `tag_type` is set to `service`,
               otherwise it is inferred from the authorization IAM token.
        :param str tag_type: (optional) The type of the tag. Supported values are
               `user`, `service` and `access`. `service` and `access` are not supported
               for IMS resources (`providers` parameter set to `ims`).
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `DeleteTagResults` object
        """

        if not tag_name:
            raise ValueError('tag_name must be provided')
        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='delete_tag',
        )
        headers.update(sdk_headers)

        params = {
            'providers': convert_list(providers),
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
        }

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        path_param_keys = ['tag_name']
        path_param_values = self.encode_path_vars(tag_name)
        path_param_dict = dict(zip(path_param_keys, path_param_values))
        url = '/v3/tags/{tag_name}'.format(**path_param_dict)
        request = self.prepare_request(
            method='DELETE',
            url=url,
            headers=headers,
            params=params,
        )

        response = self.send(request, **kwargs)
        return response

    def attach_tag(
        self,
        resources: List['Resource'],
        *,
        tag_name: str = None,
        tag_names: List[str] = None,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        impersonate_user: str = None,
        account_id: str = None,
        tag_type: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Attach tags.

        Attaches one or more tags to one or more resources. Each resource can have no more
        than 1000 tags per each 'user' and 'service' type, and no more than 250 'access'
        tags (which is the account limit).

        :param List[Resource] resources: List of resources on which the tag or tags
               are attached.
        :param str tag_name: (optional) The name of the tag to attach.
        :param List[str] tag_names: (optional) An array of tag names to attach.
        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param str impersonate_user: (optional) The user on whose behalf the attach
               operation must be performed (_for administrators only_).
        :param str account_id: (optional) The ID of the billing account of the
               tagged resource. It is a required parameter if `tag_type` is set to
               `service`. Otherwise, it is inferred from the authorization IAM token.
        :param str tag_type: (optional) The type of the tag. Supported values are
               `user`, `service` and `access`. `service` and `access` are not supported
               for IMS resources.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `TagResults` object
        """

        if resources is None:
            raise ValueError('resources must be provided')
        resources = [convert_model(x) for x in resources]
        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='attach_tag',
        )
        headers.update(sdk_headers)

        params = {
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
        }

        data = {
            'resources': resources,
            'tag_name': tag_name,
            'tag_names': tag_names,
        }
        data = {k: v for (k, v) in data.items() if v is not None}
        data = json.dumps(data)
        headers['content-type'] = 'application/json'

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v3/tags/attach'
        request = self.prepare_request(
            method='POST',
            url=url,
            headers=headers,
            params=params,
            data=data,
        )

        response = self.send(request, **kwargs)
        return response

    def detach_tag(
        self,
        resources: List['Resource'],
        *,
        tag_name: str = None,
        tag_names: List[str] = None,
        x_request_id: str = None,
        x_correlation_id: str = None,
        transaction_id: str = None,
        impersonate_user: str = None,
        account_id: str = None,
        tag_type: str = None,
        **kwargs,
    ) -> DetailedResponse:
        """
        Detach tags.

        Detaches one or more tags from one or more resources.

        :param List[Resource] resources: List of resources on which the tag or tags
               are detached.
        :param str tag_name: (optional) The name of the tag to detach.
        :param List[str] tag_names: (optional) An array of tag names to detach.
        :param str x_request_id: (optional) An alphanumeric string that is used to
               trace the request. The value  may include ASCII alphanumerics and any of
               following segment separators: space ( ), comma (,), hyphen, (-), and
               underscore (_) and may have a length up to 1024 bytes. The value is
               considered invalid and must be ignored if that value includes any other
               character or is longer than 1024 bytes or is fewer than 8 characters. If
               not specified or invalid, it is automatically replaced by a random (version
               4) UUID.
        :param str x_correlation_id: (optional) An alphanumeric string that is used
               to trace the request as a part of a larger context: the same value is used
               for downstream requests and retries of those requests. The value may
               include ASCII alphanumerics and any of following segment separators: space
               ( ), comma (,), hyphen, (-), and underscore (_) and may have a length up to
               1024 bytes. The value is considered invalid and must be ignored if that
               value includes any other character or is longer than 1024 bytes or is fewer
               than 8 characters. If not specified or invalid, it is automatically
               replaced by a random (version 4) UUID.
        :param str transaction_id: (optional) Deprecated: An alphanumeric string
               that can be used to trace a request across services. If not specified, it
               automatically generated with the prefix "gst-".
        :param str impersonate_user: (optional) The user on whose behalf the detach
               operation must be performed (_for administrators only_).
        :param str account_id: (optional) The ID of the billing account of the
               untagged resource. It is a required parameter if `tag_type` is set to
               `service`, otherwise it is inferred from the authorization IAM token.
        :param str tag_type: (optional) The type of the tag. Supported values are
               `user`, `service` and `access`. `service` and `access` are not supported
               for IMS resources.
        :param dict headers: A `dict` containing the request headers
        :return: A `DetailedResponse` containing the result, headers and HTTP status code.
        :rtype: DetailedResponse with `dict` result representing a `TagResults` object
        """

        if resources is None:
            raise ValueError('resources must be provided')
        resources = [convert_model(x) for x in resources]
        headers = {
            'x-request-id': x_request_id,
            'x-correlation-id': x_correlation_id,
            'transaction-id': transaction_id,
        }
        sdk_headers = get_sdk_headers(
            service_name=self.DEFAULT_SERVICE_NAME,
            service_version='V1',
            operation_id='detach_tag',
        )
        headers.update(sdk_headers)

        params = {
            'impersonate_user': impersonate_user,
            'account_id': account_id,
            'tag_type': tag_type,
        }

        data = {
            'resources': resources,
            'tag_name': tag_name,
            'tag_names': tag_names,
        }
        data = {k: v for (k, v) in data.items() if v is not None}
        data = json.dumps(data)
        headers['content-type'] = 'application/json'

        if 'headers' in kwargs:
            headers.update(kwargs.get('headers'))
            del kwargs['headers']
        headers['Accept'] = 'application/json'

        url = '/v3/tags/detach'
        request = self.prepare_request(
            method='POST',
            url=url,
            headers=headers,
            params=params,
            data=data,
        )

        response = self.send(request, **kwargs)
        return response


class ListTagsEnums:
    """
    Enums for list_tags parameters.
    """

    class TagType(str, Enum):
        """
        The type of the tag you want to list. Supported values are `user`, `service` and
        `access`.
        """

        USER = 'user'
        SERVICE = 'service'
        ACCESS = 'access'

    class Providers(str, Enum):
        """
        Select a provider. Supported values are `ghost` and `ims`. To list both Global
        Search and Tagging tags and infrastructure tags, use `ghost,ims`. `service` and
        `access` tags can only be attached to resources that are onboarded to Global
        Search and Tagging, so you should not set this parameter to list them.
        """

        GHOST = 'ghost'
        IMS = 'ims'

    class OrderByName(str, Enum):
        """
        Order the output by tag name.
        """

        ASC = 'asc'
        DESC = 'desc'


class CreateTagEnums:
    """
    Enums for create_tag parameters.
    """

    class TagType(str, Enum):
        """
        The type of the tags you want to create. The only allowed value is `access`.
        """

        ACCESS = 'access'


class DeleteTagAllEnums:
    """
    Enums for delete_tag_all parameters.
    """

    class Providers(str, Enum):
        """
        Select a provider. Supported values are `ghost` and `ims`.
        """

        GHOST = 'ghost'
        IMS = 'ims'

    class TagType(str, Enum):
        """
        The type of the tag. Supported values are `user`, `service` and `access`.
        `service` and `access` are not supported for IMS resources (`providers` parameter
        set to `ims`).
        """

        USER = 'user'
        SERVICE = 'service'
        ACCESS = 'access'


class DeleteTagEnums:
    """
    Enums for delete_tag parameters.
    """

    class Providers(str, Enum):
        """
        Select a provider. Supported values are `ghost` and `ims`. To delete tags both in
        Global Search and Tagging and in IMS, use `ghost,ims`.
        """

        GHOST = 'ghost'
        IMS = 'ims'

    class TagType(str, Enum):
        """
        The type of the tag. Supported values are `user`, `service` and `access`.
        `service` and `access` are not supported for IMS resources (`providers` parameter
        set to `ims`).
        """

        USER = 'user'
        SERVICE = 'service'
        ACCESS = 'access'


class AttachTagEnums:
    """
    Enums for attach_tag parameters.
    """

    class TagType(str, Enum):
        """
        The type of the tag. Supported values are `user`, `service` and `access`.
        `service` and `access` are not supported for IMS resources.
        """

        USER = 'user'
        SERVICE = 'service'
        ACCESS = 'access'


class DetachTagEnums:
    """
    Enums for detach_tag parameters.
    """

    class TagType(str, Enum):
        """
        The type of the tag. Supported values are `user`, `service` and `access`.
        `service` and `access` are not supported for IMS resources.
        """

        USER = 'user'
        SERVICE = 'service'
        ACCESS = 'access'


##############################################################################
# Models
##############################################################################


class CreateTagResults:
    """
    Results of a create tag(s) request.

    :attr List[CreateTagResultsResultsItem] results: (optional) Array of results of
          a create_tag request.
    """

    def __init__(
        self,
        *,
        results: List['CreateTagResultsResultsItem'] = None,
    ) -> None:
        """
        Initialize a CreateTagResults object.

        :param List[CreateTagResultsResultsItem] results: (optional) Array of
               results of a create_tag request.
        """
        self.results = results

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'CreateTagResults':
        """Initialize a CreateTagResults object from a json dictionary."""
        args = {}
        if 'results' in _dict:
            args['results'] = [CreateTagResultsResultsItem.from_dict(v) for v in _dict.get('results')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a CreateTagResults object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'results') and self.results is not None:
            results_list = []
            for v in self.results:
                if isinstance(v, dict):
                    results_list.append(v)
                else:
                    results_list.append(v.to_dict())
            _dict['results'] = results_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this CreateTagResults object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'CreateTagResults') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'CreateTagResults') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class CreateTagResultsResultsItem:
    """
    CreateTagResultsResultsItem.

    :attr str tag_name: (optional) The name of the tag created.
    :attr bool is_error: (optional) true if the tag was not created (for example,
          the tag already exists).
    """

    def __init__(
        self,
        *,
        tag_name: str = None,
        is_error: bool = None,
    ) -> None:
        """
        Initialize a CreateTagResultsResultsItem object.

        :param str tag_name: (optional) The name of the tag created.
        :param bool is_error: (optional) true if the tag was not created (for
               example, the tag already exists).
        """
        self.tag_name = tag_name
        self.is_error = is_error

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'CreateTagResultsResultsItem':
        """Initialize a CreateTagResultsResultsItem object from a json dictionary."""
        args = {}
        if 'tag_name' in _dict:
            args['tag_name'] = _dict.get('tag_name')
        if 'is_error' in _dict:
            args['is_error'] = _dict.get('is_error')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a CreateTagResultsResultsItem object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'tag_name') and self.tag_name is not None:
            _dict['tag_name'] = self.tag_name
        if hasattr(self, 'is_error') and self.is_error is not None:
            _dict['is_error'] = self.is_error
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this CreateTagResultsResultsItem object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'CreateTagResultsResultsItem') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'CreateTagResultsResultsItem') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class DeleteTagResults:
    """
    Results of a delete_tag request.

    :attr List[DeleteTagResultsItem] results: (optional) Array of results of a
          delete_tag request.
    """

    def __init__(
        self,
        *,
        results: List['DeleteTagResultsItem'] = None,
    ) -> None:
        """
        Initialize a DeleteTagResults object.

        :param List[DeleteTagResultsItem] results: (optional) Array of results of a
               delete_tag request.
        """
        self.results = results

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'DeleteTagResults':
        """Initialize a DeleteTagResults object from a json dictionary."""
        args = {}
        if 'results' in _dict:
            args['results'] = [DeleteTagResultsItem.from_dict(v) for v in _dict.get('results')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a DeleteTagResults object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'results') and self.results is not None:
            results_list = []
            for v in self.results:
                if isinstance(v, dict):
                    results_list.append(v)
                else:
                    results_list.append(v.to_dict())
            _dict['results'] = results_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this DeleteTagResults object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'DeleteTagResults') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'DeleteTagResults') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class DeleteTagResultsItem:
    """
    Result of a delete_tag request.

    :attr str provider: (optional) The provider of the tag.
    :attr bool is_error: (optional) It is `true` if the operation exits with an
          error (for example, the tag does not exist).
    """

    # The set of defined properties for the class
    _properties = frozenset(['provider', 'is_error'])

    def __init__(
        self,
        *,
        provider: str = None,
        is_error: bool = None,
        **kwargs,
    ) -> None:
        """
        Initialize a DeleteTagResultsItem object.

        :param str provider: (optional) The provider of the tag.
        :param bool is_error: (optional) It is `true` if the operation exits with
               an error (for example, the tag does not exist).
        :param **kwargs: (optional) Any additional properties.
        """
        self.provider = provider
        self.is_error = is_error
        for _key, _value in kwargs.items():
            setattr(self, _key, _value)

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'DeleteTagResultsItem':
        """Initialize a DeleteTagResultsItem object from a json dictionary."""
        args = {}
        if 'provider' in _dict:
            args['provider'] = _dict.get('provider')
        if 'is_error' in _dict:
            args['is_error'] = _dict.get('is_error')
        args.update({k: v for (k, v) in _dict.items() if k not in cls._properties})
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a DeleteTagResultsItem object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'provider') and self.provider is not None:
            _dict['provider'] = self.provider
        if hasattr(self, 'is_error') and self.is_error is not None:
            _dict['is_error'] = self.is_error
        for _key in [k for k in vars(self).keys() if k not in DeleteTagResultsItem._properties]:
            _dict[_key] = getattr(self, _key)
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def get_properties(self) -> Dict:
        """Return a dictionary of arbitrary properties from this instance of DeleteTagResultsItem"""
        _dict = {}

        for _key in [k for k in vars(self).keys() if k not in DeleteTagResultsItem._properties]:
            _dict[_key] = getattr(self, _key)
        return _dict

    def set_properties(self, _dict: dict):
        """Set a dictionary of arbitrary properties to this instance of DeleteTagResultsItem"""
        for _key in [k for k in vars(self).keys() if k not in DeleteTagResultsItem._properties]:
            delattr(self, _key)

        for _key, _value in _dict.items():
            if _key not in DeleteTagResultsItem._properties:
                setattr(self, _key, _value)

    def __str__(self) -> str:
        """Return a `str` version of this DeleteTagResultsItem object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'DeleteTagResultsItem') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'DeleteTagResultsItem') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other

    class ProviderEnum(str, Enum):
        """
        The provider of the tag.
        """

        GHOST = 'ghost'
        IMS = 'ims'


class DeleteTagsResult:
    """
    Results of deleting unattatched tags.

    :attr int total_count: (optional) The number of tags that have been deleted.
    :attr bool errors: (optional) It is set to true if there is at least one tag
          operation in error.
    :attr List[DeleteTagsResultItem] items: (optional) The list of tag operation
          results.
    """

    def __init__(
        self,
        *,
        total_count: int = None,
        errors: bool = None,
        items: List['DeleteTagsResultItem'] = None,
    ) -> None:
        """
        Initialize a DeleteTagsResult object.

        :param int total_count: (optional) The number of tags that have been
               deleted.
        :param bool errors: (optional) It is set to true if there is at least one
               tag operation in error.
        :param List[DeleteTagsResultItem] items: (optional) The list of tag
               operation results.
        """
        self.total_count = total_count
        self.errors = errors
        self.items = items

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'DeleteTagsResult':
        """Initialize a DeleteTagsResult object from a json dictionary."""
        args = {}
        if 'total_count' in _dict:
            args['total_count'] = _dict.get('total_count')
        if 'errors' in _dict:
            args['errors'] = _dict.get('errors')
        if 'items' in _dict:
            args['items'] = [DeleteTagsResultItem.from_dict(v) for v in _dict.get('items')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a DeleteTagsResult object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'total_count') and self.total_count is not None:
            _dict['total_count'] = self.total_count
        if hasattr(self, 'errors') and self.errors is not None:
            _dict['errors'] = self.errors
        if hasattr(self, 'items') and self.items is not None:
            items_list = []
            for v in self.items:
                if isinstance(v, dict):
                    items_list.append(v)
                else:
                    items_list.append(v.to_dict())
            _dict['items'] = items_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this DeleteTagsResult object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'DeleteTagsResult') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'DeleteTagsResult') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class DeleteTagsResultItem:
    """
    Result of a delete_tags request.

    :attr str tag_name: (optional) The name of the deleted tag.
    :attr bool is_error: (optional) true if the tag was not deleted.
    """

    def __init__(
        self,
        *,
        tag_name: str = None,
        is_error: bool = None,
    ) -> None:
        """
        Initialize a DeleteTagsResultItem object.

        :param str tag_name: (optional) The name of the deleted tag.
        :param bool is_error: (optional) true if the tag was not deleted.
        """
        self.tag_name = tag_name
        self.is_error = is_error

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'DeleteTagsResultItem':
        """Initialize a DeleteTagsResultItem object from a json dictionary."""
        args = {}
        if 'tag_name' in _dict:
            args['tag_name'] = _dict.get('tag_name')
        if 'is_error' in _dict:
            args['is_error'] = _dict.get('is_error')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a DeleteTagsResultItem object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'tag_name') and self.tag_name is not None:
            _dict['tag_name'] = self.tag_name
        if hasattr(self, 'is_error') and self.is_error is not None:
            _dict['is_error'] = self.is_error
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this DeleteTagsResultItem object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'DeleteTagsResultItem') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'DeleteTagsResultItem') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class Resource:
    """
    A resource that might have tags that are attached.

    :attr str resource_id: The CRN or IMS ID of the resource.
    :attr str resource_type: (optional) The IMS resource type of the resource.
    """

    def __init__(
        self,
        resource_id: str,
        *,
        resource_type: str = None,
    ) -> None:
        """
        Initialize a Resource object.

        :param str resource_id: The CRN or IMS ID of the resource.
        :param str resource_type: (optional) The IMS resource type of the resource.
        """
        self.resource_id = resource_id
        self.resource_type = resource_type

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'Resource':
        """Initialize a Resource object from a json dictionary."""
        args = {}
        if 'resource_id' in _dict:
            args['resource_id'] = _dict.get('resource_id')
        else:
            raise ValueError('Required property \'resource_id\' not present in Resource JSON')
        if 'resource_type' in _dict:
            args['resource_type'] = _dict.get('resource_type')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a Resource object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'resource_id') and self.resource_id is not None:
            _dict['resource_id'] = self.resource_id
        if hasattr(self, 'resource_type') and self.resource_type is not None:
            _dict['resource_type'] = self.resource_type
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this Resource object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'Resource') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'Resource') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class Tag:
    """
    A tag.

    :attr str name: The name of the tag.
    """

    def __init__(
        self,
        name: str,
    ) -> None:
        """
        Initialize a Tag object.

        :param str name: The name of the tag.
        """
        self.name = name

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'Tag':
        """Initialize a Tag object from a json dictionary."""
        args = {}
        if 'name' in _dict:
            args['name'] = _dict.get('name')
        else:
            raise ValueError('Required property \'name\' not present in Tag JSON')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a Tag object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'name') and self.name is not None:
            _dict['name'] = self.name
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this Tag object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'Tag') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'Tag') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class TagList:
    """
    A list of tags.

    :attr int total_count: (optional) Set the occurrences of the total tags that are
          associated with this account.
    :attr int offset: (optional) The offset at which tags are returned.
    :attr int limit: (optional) The number of tags requested to be returned.
    :attr List[Tag] items: (optional) Array of output results.
    """

    def __init__(
        self,
        *,
        total_count: int = None,
        offset: int = None,
        limit: int = None,
        items: List['Tag'] = None,
    ) -> None:
        """
        Initialize a TagList object.

        :param int total_count: (optional) Set the occurrences of the total tags
               that are associated with this account.
        :param int offset: (optional) The offset at which tags are returned.
        :param int limit: (optional) The number of tags requested to be returned.
        :param List[Tag] items: (optional) Array of output results.
        """
        self.total_count = total_count
        self.offset = offset
        self.limit = limit
        self.items = items

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'TagList':
        """Initialize a TagList object from a json dictionary."""
        args = {}
        if 'total_count' in _dict:
            args['total_count'] = _dict.get('total_count')
        if 'offset' in _dict:
            args['offset'] = _dict.get('offset')
        if 'limit' in _dict:
            args['limit'] = _dict.get('limit')
        if 'items' in _dict:
            args['items'] = [Tag.from_dict(v) for v in _dict.get('items')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a TagList object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'total_count') and self.total_count is not None:
            _dict['total_count'] = self.total_count
        if hasattr(self, 'offset') and self.offset is not None:
            _dict['offset'] = self.offset
        if hasattr(self, 'limit') and self.limit is not None:
            _dict['limit'] = self.limit
        if hasattr(self, 'items') and self.items is not None:
            items_list = []
            for v in self.items:
                if isinstance(v, dict):
                    items_list.append(v)
                else:
                    items_list.append(v.to_dict())
            _dict['items'] = items_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this TagList object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'TagList') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'TagList') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class TagResults:
    """
    Results of an attach_tag or detach_tag request.

    :attr List[TagResultsItem] results: (optional) Array of results of an attach_tag
          or detach_tag request.
    """

    def __init__(
        self,
        *,
        results: List['TagResultsItem'] = None,
    ) -> None:
        """
        Initialize a TagResults object.

        :param List[TagResultsItem] results: (optional) Array of results of an
               attach_tag or detach_tag request.
        """
        self.results = results

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'TagResults':
        """Initialize a TagResults object from a json dictionary."""
        args = {}
        if 'results' in _dict:
            args['results'] = [TagResultsItem.from_dict(v) for v in _dict.get('results')]
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a TagResults object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'results') and self.results is not None:
            results_list = []
            for v in self.results:
                if isinstance(v, dict):
                    results_list.append(v)
                else:
                    results_list.append(v.to_dict())
            _dict['results'] = results_list
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this TagResults object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'TagResults') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'TagResults') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other


class TagResultsItem:
    """
    Result of an attach_tag or detach_tag request for a tagged resource.

    :attr str resource_id: The CRN or IMS ID of the resource.
    :attr bool is_error: (optional) It is `true` if the operation exits with an
          error.
    """

    def __init__(
        self,
        resource_id: str,
        *,
        is_error: bool = None,
    ) -> None:
        """
        Initialize a TagResultsItem object.

        :param str resource_id: The CRN or IMS ID of the resource.
        :param bool is_error: (optional) It is `true` if the operation exits with
               an error.
        """
        self.resource_id = resource_id
        self.is_error = is_error

    @classmethod
    def from_dict(cls, _dict: Dict) -> 'TagResultsItem':
        """Initialize a TagResultsItem object from a json dictionary."""
        args = {}
        if 'resource_id' in _dict:
            args['resource_id'] = _dict.get('resource_id')
        else:
            raise ValueError('Required property \'resource_id\' not present in TagResultsItem JSON')
        if 'is_error' in _dict:
            args['is_error'] = _dict.get('is_error')
        return cls(**args)

    @classmethod
    def _from_dict(cls, _dict):
        """Initialize a TagResultsItem object from a json dictionary."""
        return cls.from_dict(_dict)

    def to_dict(self) -> Dict:
        """Return a json dictionary representing this model."""
        _dict = {}
        if hasattr(self, 'resource_id') and self.resource_id is not None:
            _dict['resource_id'] = self.resource_id
        if hasattr(self, 'is_error') and self.is_error is not None:
            _dict['is_error'] = self.is_error
        return _dict

    def _to_dict(self):
        """Return a json dictionary representing this model."""
        return self.to_dict()

    def __str__(self) -> str:
        """Return a `str` version of this TagResultsItem object."""
        return json.dumps(self.to_dict(), indent=2)

    def __eq__(self, other: 'TagResultsItem') -> bool:
        """Return `true` when self and other are equal, false otherwise."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: 'TagResultsItem') -> bool:
        """Return `true` when self and other are not equal, false otherwise."""
        return not self == other
